from .agent import JAImsAgent

from .default_tool_manager import JAImsDefaultToolManager

from .default_history_manager import (
    JAImsDefaultHistoryManager,
    JAImsLastNHistoryOptimizer,
)
from .entities import (
    JAImsMessage,
    JAImsImageContent,
    JAImsContentType,
    JAImsMessageRole,
    JAImsToolResponse,
    JAImsFunctionTool,
    JAImsMaxConsecutiveFunctionCallsExceeded,
    JAImsFunctionToolDescriptor,
    JAImsToolCall,
    JAImsLLMConfig,
    JAImsOptions,
    JAImsImageContentType,
    JAImsMaxRetriesExceeded,
    JAImsModelCode,
    JAImsStreamingMessage,
    JAImsUnexpectedFunctionCall,
)

from .function_tool_decorator import jaimsfunctiontool

from .interfaces import JAImsToolManager, JAImsHistoryManager, JAImsLLMInterface

from pydantic import BaseModel, Field, create_model

__all__ = [
    "JAImsAgent",
    "JAImsDefaultToolManager",
    "JAImsDefaultHistoryManager",
    "JAImsLastNHistoryOptimizer",
    "JAImsMessage",
    "JAImsImageContent",
    "JAImsContentType",
    "JAImsMessageRole",
    "JAImsToolResponse",
    "JAImsFunctionTool",
    "JAImsMaxConsecutiveFunctionCallsExceeded",
    "JAImsFunctionToolDescriptor",
    "JAImsToolCall",
    "JAImsLLMConfig",
    "JAImsOptions",
    "JAImsImageContentType",
    "JAImsMaxRetriesExceeded",
    "JAImsModelCode",
    "JAImsStreamingMessage",
    "JAImsUnexpectedFunctionCall",
    "JAImsToolManager",
    "JAImsHistoryManager",
    "JAImsLLMInterface",
    "jaimsfunctiontool",
    "BaseModel",
    "Field",
    "create_model",
]



import os
from ...interfaces import JAImsLLMInterface
from ...entities import (
    JAImsMessage,
    JAImsFunctionTool,
    JAImsToolCall,
    JAImsMessageRole,
    JAImsImageContent,
    JAImsStreamingMessage,
    JAImsOptions,
)
from ..shared.image_utilities import image_to_bytes
from ..shared.exponential_backoff_operation import (
    call_with_exponential_backoff,
    ErrorHandlingMethod,
)

from typing import Iterable, List, Generator, Optional
from PIL import Image


from google.generativeai import GenerativeModel
from google.generativeai.types import content_types
from google.generativeai.types import generation_types
import google.ai.generativelanguage as glm
import jsonref


class JAImsGoogleGenerativeAIAdapter(JAImsLLMInterface):
    def __init__(
        self,
        model: str,
        generation_config: Optional[generation_types.GenerationConfigType] = None,
        options: Optional[JAImsOptions] = None,
        api_key: Optional[str] = None,
    ):
        self.api_key = api_key or os.getenv("GOOGLE_API_KEY")
        if not self.api_key:
            raise Exception("GOOGLE_API_KEY not provided.")

        self.model = model
        self.generation_config = generation_config
        self.options = options or JAImsOptions()

    def call(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
    ) -> JAImsMessage:
        response = self.__get_gemini_response(messages, tools, tool_constraints)
        assert isinstance(response, generation_types.GenerateContentResponse)
        return self.__gemini_to_jaims_message(response)

    def call_streaming(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
    ) -> Generator[JAImsStreamingMessage, None, None]:
        response = self.__get_gemini_response(
            messages, tools, stream=True, tool_constraints=tool_constraints
        )
        assert isinstance(response, Iterable)
        for r in response:
            if r.candidates:
                message = self.__gemini_to_jaims_message(r)
                yield JAImsStreamingMessage(
                    textDelta=message.get_text() or None,
                    message=message,
                )

    def __jaims_role_to_gemini(self, role: JAImsMessageRole) -> str:
        if role == JAImsMessageRole.USER:
            return "user"
        else:
            return "model"

    def __jaims_tools_to_gemini(
        self, jaims_tools: List[JAImsFunctionTool]
    ) -> List[content_types.Tool]:
        function_declarations = []
        for jaims_tool in jaims_tools:

            function_declarations.append(
                content_types.FunctionDeclaration(
                    name=jaims_tool.descriptor.name,
                    description=jaims_tool.descriptor.description,
                    parameters=jaims_tool.descriptor.json_schema(
                        remove_any_of=True, dereference=True
                    ),
                ),
            )

        return [content_types.Tool(function_declarations=function_declarations)]

    def __jaims_messages_to_gemini(
        self, jaims_messages: List[JAImsMessage]
    ) -> tuple[List[str], List[content_types.ContentsType]]:
        gemini_messages = []
        system_instruction = []

        for jaims_message in jaims_messages:
            if jaims_message.role == JAImsMessageRole.SYSTEM:
                system_instruction.append(jaims_message.get_text())
                continue

            role = self.__jaims_role_to_gemini(jaims_message.role)
            parts = []

            if jaims_message.contents:
                for content in jaims_message.contents:
                    if isinstance(content, str):
                        parts.append(glm.Part(text=content))
                    if isinstance(content, JAImsImageContent):
                        if isinstance(content.image, str):
                            raise ValueError("Image URL not supported")
                        elif isinstance(content.image, Image.Image):
                            mime, data = image_to_bytes(content.image)
                            parts.append(
                                glm.Part(
                                    inline_data=glm.Blob(mime_type=mime, data=data)
                                )
                            )
                        else:
                            raise ValueError("Invalid image content type")

            if jaims_message.tool_response:
                parts.append(
                    glm.Part(
                        function_response=glm.FunctionResponse(
                            name=jaims_message.tool_response.tool_name,
                            response={"result": jaims_message.tool_response.response},
                        )
                    )
                )

            if jaims_message.tool_calls:
                for tool_call in jaims_message.tool_calls:
                    parts.append(
                        glm.Part(
                            function_call=glm.FunctionCall(
                                name=tool_call.tool_name,
                                args=tool_call.tool_args,
                            )
                        )
                    )

            gemini_messages.append(
                glm.Content(
                    role=role,
                    parts=parts,
                )
            )

        return system_instruction, gemini_messages

    def __gemini_to_jaims_message(
        self, response: generation_types.GenerateContentResponse
    ) -> JAImsMessage:
        assert response.candidates is not None

        role = JAImsMessageRole.ASSISTANT
        tool_calls = []
        contents = []
        for part in response.parts:
            if fn := part.function_call:
                args = {}
                for k, v in fn.args.items():
                    args[k] = v
                tool_calls.append(
                    JAImsToolCall(
                        id=fn.name,
                        tool_name=fn.name,
                        tool_args=args,
                    )
                )

            if txt := part.text:
                contents.append(txt)

        return JAImsMessage(
            role=role,
            contents=contents,
            tool_calls=tool_calls,
            raw=response,
        )

    def __get_gemini_response(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
        stream: bool = False,
    ):

        def handle_gemini_error(error):

            # From: https://ai.google.dev/gemini-api/docs/troubleshooting
            # 400	INVALID_ARGUMENT	The request body is malformed.	Check the API reference for request format, examples, and supported versions. Using features from a newer API version with an older endpoint can cause errors.
            # 403	PERMISSION_DENIED	Your API key doesn't have the required permissions.	Check that your API key is set and has the right access.
            # 404	NOT_FOUND	The requested resource wasn't found.	Check if all parameters in your request are valid for your API version.
            # 429	RESOURCE_EXHAUSTED	You've exceeded the rate limit.	Ensure you're within the model's rate limit. Request a quota increase if needed.
            # 500	INTERNAL	An unexpected error occurred on Google's side.	Wait a bit and retry your request. If the issue persists after retrying, please report it using the Send feedback button in Google AI Studio.
            # 503	UNAVAILABLE	The service may be temporarily overloaded or down.	Wait a bit and retry your request. If the issue persists after retrying, please report it using the Send feedback button in Google AI Studio.

            if error is ValueError or not hasattr(error, "code"):
                return ErrorHandlingMethod.FAIL

            if error.code == 429 or error.code == 503 or error.code == 500:
                return ErrorHandlingMethod.EXPONENTIAL_BACKOFF

            return ErrorHandlingMethod.FAIL

        def call_gemini():
            system_instruction, gemini_messages = self.__jaims_messages_to_gemini(
                messages
            )
            gemini_tools = self.__jaims_tools_to_gemini(tools) if tools else None

            tool_config = None
            if tool_constraints and tools:
                tool_config = content_types.to_tool_config(
                    {
                        "function_calling_config": {
                            "mode": "any",
                            "allowed_function_names": tool_constraints,
                        }
                    }  # type: ignore
                )

            multimodal_model = GenerativeModel(
                self.model,
                generation_config=self.generation_config,
                tools=gemini_tools,
                tool_config=tool_config,
                system_instruction=system_instruction or None,
            )

            return multimodal_model.generate_content(
                contents=gemini_messages,
                stream=stream,
            )

        return call_with_exponential_backoff(
            call_gemini, handle_gemini_error, self.options
        )



from jaims.entities import (
    JAImsFunctionTool,
    JAImsFunctionToolDescriptor,
)
from typing import Any, Optional, Dict
import inspect
from pydantic import BaseModel, Field, create_model


def jaimsfunctiontool(
    name: Optional[str] = None,
    description: Optional[str] = None,
    params_descriptions: Optional[Dict[str, str]] = None,
    return_value_object_name: str = "ToolResponse",
):
    """
    Decorator function for creating JAIms function tools.

    This is an experimental feature and I'm planning to improve it using reflection and other techniques to better infer the types of the parameters and return values.
    Right now the decorated function MUST have type annotations in order for the decorator to work, and the decorator supports only simple python types and pydantic models.

    When the decorated function expects only a pydantic model, it is forwarded as-is to the LLM (parsed with its json schema), when the function expects multiple parameters, The decorator creates a new pydantic model by inspecting the function signature.
    You can customize the name of the generated model by passing the return_value_object_name parameter (defaults to "ToolResponse")

    Args:
        name (Optional[str]): The name of the tool. If not provided, the name of the decorated function will be used.
        description (Optional[str]): The description of the tool. If not provided, an empty string will be used.
        params_descriptions (Optional[Dict[str, str]]): A dictionary mapping parameter names to their descriptions.
            If not provided, fields will have an empty string as description.
        return_value_object_name (str): The name of the return value object. Defaults to "ToolResponse". Used only when the function expects multiple parameters.

    Returns:
        The decorated function as a JAImsFunctionTool.

    Raises:
        ValueError: If a parameter has no type annotation.

    Example usage:

    ```python
        @jaimsfunctiontool(name="MyTool", description="This is my tool", params_descriptions={"param1": "The first parameter", "param2": "The second parameter"})
        def my_function(param1: int, param2: str = "default"):
            return param1 + len(param2)
    ```
    """

    def decorator(func):
        tool_name = name if name else func.__name__
        tool_description = description or ""
        params_models = {}
        sig = inspect.signature(func)
        has_base_model = False

        for param_name, param in sig.parameters.items():

            # skip self, cls and return
            if param_name in ["self", "cls", "return"]:
                continue

            if param.annotation is inspect.Parameter.empty:
                raise ValueError(
                    f"Parameter '{param_name}' has no type annotation, please make sure you're type hinting every parameter"
                )

            param_description = (
                params_descriptions.get(param_name, "") if params_descriptions else ""
            )

            if param.default is inspect._empty:
                params_models[param_name] = (
                    param.annotation,
                    Field(description=param_description),
                )
            else:
                # check if the param is a BaseModel or a subclass of it
                if isinstance(param.annotation, type) and issubclass(
                    param.annotation, BaseModel
                ):
                    has_base_model = True

                params_models[param_name] = (
                    param.annotation,
                    Field(default=param.default, description=param_description),
                )

        if len(params_models) == 1 and has_base_model:
            output_model = list(params_models.values())[0][0]
        else:
            output_model = create_model(return_value_object_name, **params_models)

        descriptor = JAImsFunctionToolDescriptor(
            name=tool_name, description=tool_description, params=output_model
        )

        def formatter(data: Dict[str, Any]):
            formatted_model = output_model.model_validate(data)
            kwargs = {
                name: getattr(formatted_model, name) for name in params_models.keys()
            }

            return (), kwargs

        return JAImsFunctionTool(
            function=func,
            formatter=formatter,
            descriptor=descriptor,
        )

    return decorator



from typing import List, Optional
import unittest

from pydantic import BaseModel, Field
from jaims import (
    jaimsfunctiontool,
)


class TestJAIMSFunctionToolDecorator(unittest.TestCase):

    def test_decorator_sets_passed_tool_name(self):

        @jaimsfunctiontool(name="add_numbers_alt", description="Add two numbers")
        def add_numbers(a: int, b: int) -> int:
            return a + b

        self.assertEqual(add_numbers.descriptor.name, "add_numbers_alt")

    def test_decorator_sets_passed_description(self):

        @jaimsfunctiontool(name="add_numbers", description="Add two numbers alt")
        def add_numbers(a: int, b: int) -> int:
            return a + b

        self.assertEqual(add_numbers.descriptor.description, "Add two numbers alt")

    def test_decorator_sets_function_name_when_not_passed(self):

        @jaimsfunctiontool(description="Add two numbers")
        def add_numbers(a: int, b: int) -> int:
            return a + b

        self.assertEqual(add_numbers.descriptor.name, "add_numbers")

    def test_decorated_function_gets_called_by_wrapper(self):

        class MockClass(BaseModel):
            a: int = Field(description="Mock a")
            b: int = Field(description="Mock b")

        @jaimsfunctiontool(description="mock wrapped function")
        def mock_wrapped_function(
            input: str,
            number: int,
            flag: bool,
            items: List[str],
            mock_class: MockClass,
        ) -> str:
            items_str = " ".join(items)
            return f"Hello {input} {number} {flag} {items_str} {mock_class.a} {mock_class.b}"

        result = mock_wrapped_function(
            "Mondo", 42, True, ["a", "b"], MockClass(a=11, b=22)
        )
        self.assertEqual(result, "Hello Mondo 42 True a b 11 22")

    def test_decorated_function_gets_called_with_formatter(self):

        class MockClass(BaseModel):
            a: int = Field(description="Mock a")
            b: int = Field(description="Mock b")

        @jaimsfunctiontool(description="mock wrapped function")
        def mock_wrapped_function(
            input: str,
            number: int,
            flag: bool,
            items: List[str],
            mock_class: MockClass,
        ) -> str:
            items_str = " ".join(items)
            return f"Hello {input} {number} {flag} {items_str} {mock_class.a} {mock_class.b}"

        payload = {
            "input": "Mondo",
            "number": 42,
            "flag": True,
            "items": ["a", "b"],
            "mock_class": {"a": 11, "b": 22},
        }

        result = mock_wrapped_function.call_raw(**payload)
        self.assertEqual(result, "Hello Mondo 42 True a b 11 22")

    def test_decorated_tool_descriptor_param_type_is_base_model_when_passed_base_model_only(
        self,
    ):

        class MockClass(BaseModel):
            a: int = Field(description="Mock a")
            b: int = Field(description="Mock b")

        @jaimsfunctiontool(description="mock wrapped function")
        def mock_wrapped_function(mock_class: MockClass):
            return f"{mock_class.a} {mock_class.b}"

        self.assertIsInstance(mock_wrapped_function.descriptor.params, type(MockClass))

    def test_decorator_skips_return_type_in_json_schema(self):

        @jaimsfunctiontool(description="Add two numbers")
        def add_numbers(a: int, b: int) -> int:
            return a + b

        output_schema = add_numbers.descriptor.json_schema()

        self.assertNotIn("return", output_schema["properties"])

    def test_decorator_applies_descriptions_when_passed(self):

        @jaimsfunctiontool(
            name="add_numbers",
            description="Add two numbers",
            params_descriptions={"a": "First number", "b": "Second number"},
        )
        def add_numbers(a: int, b: Optional[int] = None) -> int:
            return a + (b or 1)

        output_schema = add_numbers.descriptor.json_schema()

        self.assertEqual(
            output_schema["properties"]["a"]["description"], "First number"
        )
        self.assertEqual(
            output_schema["properties"]["b"]["description"], "Second number"
        )

    def test_decorator_marks_fields_with_default_values_as_not_required(self):

        @jaimsfunctiontool(description="Add two numbers")
        def add_numbers(a: int, b: Optional[int] = None, c: int = 4) -> int:
            return a + (b or 1)

        output_schema = add_numbers.descriptor.json_schema()

        self.assertNotIn("b", output_schema["required"])
        self.assertNotIn("c", output_schema["required"])

    def test_decorator_works_on_instances(self):

        class MockClass(BaseModel):
            a: int = Field(description="Mock a")

        class MockClassMethods:
            @jaimsfunctiontool(description="mock wrapped function")
            def mock_wrapped_function(
                self, param_one: str, param_two: MockClass
            ) -> str:
                self.called = True
                return f"{param_one} {param_two.a}"

        mock_instance = MockClassMethods()
        payload = {"param_one": "Hello", "param_two": {"a": 42}}
        result = mock_instance.mock_wrapped_function.call_raw(**payload)
        self.assertTrue(mock_instance.called)
        self.assertEqual(result, "Hello 42")

    def test_decorator_works_on_functions(self):

        class MockClass(BaseModel):
            a: int = Field(description="Mock a")

        @jaimsfunctiontool(description="mock wrapped function")
        def mock_wrapped_function(param_one: str, param_two: MockClass) -> str:
            return f"{param_one} {param_two.a}"

        payload = {"param_one": "Hello", "param_two": {"a": 42}}
        result = mock_wrapped_function.call_raw(**payload)
        self.assertEqual(result, "Hello 42")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Literal

from pydantic import BaseModel

if TYPE_CHECKING:
    from .interfaces import JAImsLLMInterface, JAImsHistoryManager, JAImsToolManager

from typing import Generator, List, Optional


from jaims.default_tool_manager import JAImsDefaultToolManager


from jaims.entities import (
    JAImsFunctionToolDescriptor,
    JAImsMaxConsecutiveFunctionCallsExceeded,
    JAImsMessage,
    JAImsStreamingMessage,
    JAImsFunctionTool,
    JAImsLLMConfig,
    JAImsOptions,
)


class JAImsAgent:
    """
    Base  JAIms Agent class, interacts with the JAImsLLMInterface to run messages and tools.

    Tools can be injected in the constructor once, or passed when invoking (overriding the injected tools at every invocation).
    Provide an history_manager if you plan to use the agent as a chat agent and want to keep track of the conversation.
    Provide a tool_manager if you want to customize the tool handling.

    Args:
        llm_interface (JAImsLLMInterface): The LLM interface to use.
        tools (Optional[List[JAImsFunctionTool]]): The list of tools to use. Defaults to None.
        history_manager (Optional[JAImsHistoryManager]): The history manager to use. Defaults to None.
        tool_manager (Optional[JAImsToolManager]): The tool manager to use. Defaults to None.
        max_consecutive_tool_calls (int): The maximum number of consecutive tool calls allowed. Defaults to 10.
    """

    def __init__(
        self,
        llm_interface: JAImsLLMInterface,
        history_manager: Optional[JAImsHistoryManager] = None,
        tool_manager: Optional[JAImsToolManager] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
        max_consecutive_tool_calls: int = 10,
    ):
        self.llm_interface = llm_interface
        self.tool_manager = tool_manager or JAImsDefaultToolManager()
        self.tools = tools or []
        self.history_manager = history_manager
        self.max_consecutive_tool_calls = max_consecutive_tool_calls
        self.__session_iteration = -1
        self.__session_messages = []

    @staticmethod
    def build(
        model: str,
        provider: Literal["openai", "google"],
        api_key: Optional[str] = None,
        options: Optional[JAImsOptions] = None,
        config: Optional[JAImsLLMConfig] = None,
        history_manager: Optional[JAImsHistoryManager] = None,
        tool_manager: Optional[JAImsToolManager] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
    ) -> JAImsAgent:
        """
        Factory method to build an agent with the specified parameters.

        Currently available providers are: [openai, google]. Make sure to install the required dependencies using:

        ```bash
        pip install jaims-py[openai, google]
        ```

        The API key will be read from the default environment variables when not provided.

        Args:
            model (str): The model to use.
            provider (Literal["openai", "google"]): The provider to use.
            api_key (Optional[str]): The API key. Defaults to None.
            options (Optional[JAImsOptions]): The options. Defaults to None.
            config (Optional[JAImsLLMConfig]): The config. Defaults to None.
            history_manager (Optional[JAImsHistoryManager]): The history manager. Defaults to None.
            tool_manager (Optional[JAImsToolManager]): The tool manager. Defaults to None.
            tools (Optional[List[JAImsFunctionTool]]): The list of tools. Defaults to None.

        Returns:
            JAImsAgent: The agent instance.
        """

        assert provider in [
            "openai",
            "google",
        ], f"curretnly supported providers are: [openai, google] . If you're targeting an unsupported provider you should supply your own adapter instead."

        if provider == "openai":
            from .factories import openai_factory

            return openai_factory(
                model=model,
                api_key=api_key,
                options=options,
                config=config,
                history_manager=history_manager,
                tool_manager=tool_manager,
                tools=tools,
            )
        elif provider == "google":
            from .factories import google_factory

            return google_factory(
                model=model,
                api_key=api_key,
                options=options,
                config=config,
                history_manager=history_manager,
                tool_manager=tool_manager,
                tools=tools,
            )
        else:
            raise ValueError("Provider is not supported.")

    def __update_session(self, session_messages: List[JAImsMessage]):
        self.__session_iteration += 1
        if self.__session_iteration > self.max_consecutive_tool_calls:
            raise JAImsMaxConsecutiveFunctionCallsExceeded(self.__session_iteration)

        if self.history_manager:
            self.history_manager.add_messages(session_messages)
            self.__session_messages = self.history_manager.get_messages()
        else:
            self.__session_messages.extend(session_messages)

    def __end_session(self, messages: List[JAImsMessage]):

        if self.history_manager:
            self.history_manager.add_messages(messages)

        self.__session_iteration = -1
        self.__session_messages = []

    def __get_tool_results(
        self, message: JAImsMessage, tools: List[JAImsFunctionTool]
    ) -> List[JAImsMessage]:
        tool_results = []
        if message and message.tool_calls:
            tool_call_results = self.tool_manager.handle_tool_calls(
                self, message.tool_calls, tools
            )
            tool_results.extend(tool_call_results)

        return tool_results

    def run(
        self,
        messages: Optional[List[JAImsMessage]] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
        tool_constraints: Optional[List[str]] = None,
    ) -> JAImsMessage:
        """
        Runs the agent with the given messages and tools and returns the response message.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            tools (Optional[List[JAImsFunctionTool]]): When passed, the tools will override any tools injected in the constructor, only for this run. When None, the tools injected in the constructor will be used (if any). Defaults to None.
            tool_constraints (Optional[List[str]]): The list of tool identifiers that should be used. When None, the LLM works in agent mode (if supported) and picks the tools to use. Defaults to None.

        Returns:
            JAImsMessage: The response message.
        """

        self.__update_session(messages or [])

        run_tools = tools or self.tools

        response_message = self.llm_interface.call(
            self.__session_messages, run_tools, tool_constraints=tool_constraints
        )

        tool_results = self.__get_tool_results(response_message, run_tools)
        if tool_results and not tool_constraints:
            return self.run(
                [response_message] + tool_results,
                run_tools,
                tool_constraints=tool_constraints,
            )

        self.__end_session([response_message] + tool_results)
        return response_message

    @staticmethod
    def run_model(
        model: str,
        provider: Literal["openai", "google"],
        messages: Optional[List[JAImsMessage]] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
        tools_constraints: Optional[List[str]] = None,
        api_key: Optional[str] = None,
        options: Optional[JAImsOptions] = None,
        config: Optional[JAImsLLMConfig] = None,
        tool_manager: Optional[JAImsToolManager] = None,
    ) -> JAImsMessage:
        """
        Runs the specified model with the given parameters and returns the response message.

        Args:
            model (str): The model to use.
            provider (Literal["openai", "google"]): The provider to use.
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            tools (Optional[List[JAImsFunctionTool]]): The list of tools. Defaults to None.
            tools_constraints (Optional[List[str]]): The list of tool identifiers that should be used. Defaults to None.
            api_key (Optional[str]): The API key. Defaults to None.
            options (Optional[JAImsOptions]): The options. Defaults to None.
            config (Optional[JAImsLLMConfig]): The config. Defaults to None.
            tool_manager (Optional[JAImsToolManager]): The tool manager. Defaults to None.

        Returns:
            JAImsMessage: The response message.
        """

        agent = JAImsAgent.build(
            model=model,
            provider=provider,
            api_key=api_key,
            options=options,
            config=config,
            tool_manager=tool_manager,
            tools=tools,
        )

        return agent.run(messages=messages, tool_constraints=tools_constraints)

    def run_tool(
        self,
        descriptor: JAImsFunctionToolDescriptor,
        messages: Optional[List[JAImsMessage]] = None,
    ) -> Any:
        """
        Runs a single tool with the given messages and returns the expected response data.

        Args:
            tool (JAImsFunctionToolDescriptor): The tool to run.
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.

        Returns:
            BaseModel: The expected response data defined by the tool descriptor.
        """

        response_data = None

        def callback(response: BaseModel):
            nonlocal response_data
            response_data = response

        tool = JAImsFunctionTool(
            descriptor=descriptor,
            function=callback,
        )

        self.run(messages, [tool], tool_constraints=[descriptor.name])

        if not response_data:
            raise ValueError(f"Tool {tool.descriptor.name} did not return any data.")

        return response_data

    @staticmethod
    def run_tool_model(
        model: str,
        provider: Literal["openai", "google"],
        descriptor: JAImsFunctionToolDescriptor,
        messages: Optional[List[JAImsMessage]] = None,
        api_key: Optional[str] = None,
        options: Optional[JAImsOptions] = None,
        config: Optional[JAImsLLMConfig] = None,
        tool_manager: Optional[JAImsToolManager] = None,
    ) -> Any:
        """
        Runs a single tool with the given messages and returns the expected response data.

        Args:
            model (str): The model to use.
            provider (Literal["openai", "google"]): The provider to use.
            descriptor (JAImsFunctionToolDescriptor): The tool to run.
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            api_key (Optional[str]): The API key. Defaults to None.
            options (Optional[JAImsOptions]): The options. Defaults to None.
            config (Optional[JAImsLLMConfig]): The config. Defaults to None.
            tool_manager (Optional[JAImsToolManager]): The tool manager. Defaults to None.

        Returns:
            BaseModel: The expected response data defined by the tool descriptor.
        """

        agent = JAImsAgent.build(
            model=model,
            provider=provider,
            api_key=api_key,
            options=options,
            config=config,
            tool_manager=tool_manager,
        )

        return agent.run_tool(descriptor, messages)

    def run_stream(
        self,
        messages: Optional[List[JAImsMessage]] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
        tool_constraints: Optional[List[str]] = None,
    ) -> Generator[JAImsStreamingMessage, None, None]:
        """
        Runs the agent in streaming mode with the given messages and tools and yields the streaming response message.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            tools (Optional[List[JAImsFunctionTool]]): When passed, the tools will override any tools injected in the constructor, only for this run. When None, the tools injected in the constructor will be used (if any). Defaults to None.
            tool_constraints (Optional[List[str]]): The list of tool identifiers that should be used. When None, the LLM works in agent mode (if supported) and picks the tools to use. Defaults to None.

        Yields:
            JAImsStreamingMessage: The streaming response message.
        """

        self.__update_session(messages or [])

        run_tools = tools or self.tools

        streaming_response = self.llm_interface.call_streaming(
            self.__session_messages, run_tools, tool_constraints=tool_constraints
        )

        response_message = None
        for delta_resp in streaming_response:
            response_message = delta_resp.message
            yield delta_resp

        if not response_message:
            return

        tool_results = self.__get_tool_results(response_message, run_tools)
        if tool_results and not tool_constraints:
            yield from self.run_stream(
                [response_message] + tool_results,
                run_tools,
                tool_constraints=tool_constraints,
            )
            return

        self.__end_session([response_message] + tool_results)

    def message(
        self,
        messages: Optional[List[JAImsMessage]] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
        tool_constraints: Optional[List[str]] = None,
    ) -> Optional[str]:
        """
        Sends the messages to the agent returning the response text, in a chat-like fashion.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            tools (Optional[List[JAImsFunctionTool]]): When passed, the tools will override any tools injected in the constructor, only for this run. When None, the tools injected in the constructor will be used (if any). Defaults to None.
            tool_constraints (Optional[List[str]]): The list of tool identifiers that should be used. When None, the LLM works in agent mode (if supported) and picks the tools to use. Defaults to None.

        Returns:
            Optional[str]: The response text, or None if there is no response.
        """

        message = self.run(messages, tools, tool_constraints)

        return message.get_text() if message else None

    def message_stream(
        self,
        messages: Optional[List[JAImsMessage]] = None,
        tools: Optional[List[JAImsFunctionTool]] = None,
        tool_constraints: Optional[List[str]] = None,
    ) -> Generator[str, None, None]:
        """
        Sends the messages to the agent and streams the response text, in a chat-like fashion.

        Args:
            messages (Optional[List[JAImsMessage]]): The list of messages. Defaults to None.
            tools (Optional[List[JAImsFunctionTool]]): When passed, the tools will override any tools injected in the constructor, only for this run. When None, the tools injected in the constructor will be used (if any). Defaults to None.
            tool_constraints (Optional[List[str]]): The list of tool identifiers that should be used. When None, the LLM works in agent mode (if supported) and picks the tools to use. Defaults to None.

        Yields:
            str: The text delta of each response message.
        """

        for delta_resp in self.run_stream(messages, tools, tool_constraints):
            yield delta_resp.textDelta or ""



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .agent import JAImsAgent

from abc import ABC, abstractmethod
from typing import List, Generator, Optional


from jaims.entities import (
    JAImsMessage,
    JAImsStreamingMessage,
    JAImsToolCall,
    JAImsFunctionTool,
)


class JAImsLLMInterface(ABC):
    """
    Abstract base class for JAImsLLMInterface.

    This class defines the interface for a JAIms LLM (Language Learning Model) interface.
    Subclasses of this class must implement the `call` and `call_streaming` methods.

    Attributes:
        None

    Methods:
        call: Executes the language learning model on a list of messages and tools.
        call_streaming: Executes the language learning model on a list of messages and tools in a streaming fashion.
    """

    @abstractmethod
    def call(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
    ) -> JAImsMessage:
        """
        Executes the language learning model on a list of messages and tools.

        Args:
            messages: A list of JAImsMessage objects representing the input messages.
            tools: A list of JAImsFunctionTool objects representing the tools to be applied.
            tool_constraints: An optional list of tool constraints.

        Returns:
            A JAImsMessage object representing the output message.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def call_streaming(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
    ) -> Generator[JAImsStreamingMessage, None, None]:
        """
        Executes the language learning model on a list of messages and tools in a streaming fashion.

        Args:
            messages: A list of JAImsMessage objects representing the input messages.
            tools: A list of JAImsFunctionTool objects representing the tools to be applied.
            tool_constraints: An optional list of tool constraints.

        Yields:
            A JAImsStreamingMessage object representing the output message.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError


class JAImsHistoryOptimizer(ABC):
    """
    Abstract base class for a history optimizer supported by the Default JAImsHistoryManager.

    Override this class to implement custom history optimization logic (e.g., reducing messages based on token consideration, timestamps, etc.).
    """

    @abstractmethod
    def optimize_history(self, messages: List[JAImsMessage]) -> List[JAImsMessage]:
        """
        Optimize the given list of JAIms messages.

        Parameters:
        - messages (List[JAImsMessage]): The list of JAIms messages to optimize.

        Returns:
        - List[JAImsMessage]: The optimized list of JAIms messages.
        """
        raise NotImplementedError


class JAImsHistoryManager(ABC):
    """
    Abstract base class for managing the history of JAIms messages.

    The intended use, when implementing it is to use the add_messages method to push messages in the conversation history, and the get_messages method to retrieve the messages intended to be sent to the LLM when it is invoked. Ideally you might want to send just a slice of the conversation history to the LLM, and not the entire history, to lower token consumption.
    """

    @abstractmethod
    def add_messages(self, messages: List[JAImsMessage]):
        """
        Adds a list of JAIms messages to the history.

        Args:
            messages (List[JAImsMessage]): The list of JAIms messages to add.
        """
        raise NotImplementedError

    @abstractmethod
    def get_messages(self) -> List[JAImsMessage]:
        """
        Retrieves the list of JAIms messages intended to be sent, at a specific invocation, to the LLM.

        In other words this method should return "a slice" of the conversation history when the user sends a message.

        Returns:
            List[JAImsMessage]: The list of JAIms messages.
        """
        raise NotImplementedError


class JAImsToolManager(ABC):
    """
    Abstract base class for managing JAIms tools.

    This class defines the interface for handling tool calls in JAIms.

    Attributes:
        agent (JAImsAgent): The JAIms agent associated with the tool calls.
        tool_calls (List[JAImsToolCall]): List of tool calls.
        tools (List[JAImsFunctionTool]): List of available tools.

    Returns:
        List[JAImsMessage]: List of messages generated by the tool calls.
    """

    @abstractmethod
    def handle_tool_calls(
        self,
        agent: JAImsAgent,
        tool_calls: List[JAImsToolCall],
        tools: List[JAImsFunctionTool],
    ) -> List[JAImsMessage]:
        """
        Handle tool calls in JAIms.

        This method should be implemented by subclasses to handle the tool calls.
        The list of tool calls represents a request by the LLM to execute a set of tools, the tools are the actual function wrappers that should be executed.

        Args:
            agent (JAImsAgent): The JAIms agent.
            tool_calls (List[JAImsToolCall]): List of tool calls requested by the LLM.
            tools (List[JAImsFunctionTool]): List of available tools.

        Returns:
            List[JAImsMessage]: List of messages generated by the tool calls.
        """
        raise NotImplementedError



import unittest
from unittest.mock import MagicMock, call
from jaims.agent import JAImsAgent
from jaims.entities import (
    JAImsMessage,
    JAImsMessageRole,
)

from jaims.interfaces import JAImsLLMInterface, JAImsHistoryManager, JAImsToolManager


class TestJAImsAgent(unittest.TestCase):
    def setUp(self):
        self.llm_interface = MagicMock(spec=JAImsLLMInterface)
        self.history_manager = MagicMock(spec=JAImsHistoryManager)
        self.tool_manager = MagicMock(spec=JAImsToolManager)
        self.tools = MagicMock()
        self.mock_message = JAImsMessage(
            role=JAImsMessageRole.USER,
            contents=[
                "Hello",
            ],
        )

        self.mock_message_response = JAImsMessage(
            role=JAImsMessageRole.ASSISTANT,
            contents=[
                "World",
            ],
        )
        self.mock_message_response_tool_calls = JAImsMessage.tool_call_message(
            tool_calls=[MagicMock()]
        )
        self.mock_tools = MagicMock()

    def tearDown(self) -> None:
        self.llm_interface.reset_mock()
        self.history_manager.reset_mock()
        self.tool_manager.reset_mock()
        self.tools.reset_mock()

        return super().tearDown()

    def test_run_no_messages(self):
        self.llm_interface.call.return_value = self.mock_message_response

        sut = JAImsAgent(
            llm_interface=self.llm_interface,
        )

        response = sut.message()
        self.llm_interface.call.assert_called_once_with([], [], tool_constraints=None)
        self.assertEqual(response, "World")

    def test_run_messages_and_tools(self):
        self.llm_interface.call.return_value = self.mock_message_response

        sut = JAImsAgent(
            llm_interface=self.llm_interface,
            tools=self.mock_tools,
        )

        response = sut.message([self.mock_message], tool_constraints=["hello"])
        self.llm_interface.call.assert_called_once_with(
            [self.mock_message], self.mock_tools, tool_constraints=["hello"]
        )
        self.assertEqual(response, "World")

    def test_run_passes_messages_to_history_maager(self):
        mock_history_return = MagicMock()
        self.llm_interface.call.return_value = self.mock_message_response
        self.history_manager.get_messages.return_value = mock_history_return

        sut = JAImsAgent(
            llm_interface=self.llm_interface,
            history_manager=self.history_manager,
            tools=self.mock_tools,
        )

        response = sut.message([self.mock_message])
        self.assertEqual(response, "World")

        self.llm_interface.call.assert_called_once_with(
            mock_history_return, self.mock_tools, tool_constraints=None
        )

        self.history_manager.assert_has_calls(
            [
                call.add_messages([self.mock_message]),
                call.get_messages(),
                call.add_messages([self.mock_message_response]),
            ]
        )

    def test_run_passes_tool_calls_to_tool_manager_when_receiving_tool_calls(self):
        first_message_mock = MagicMock()
        tool_response_mock = MagicMock()
        self.llm_interface.call.side_effect = [
            self.mock_message_response_tool_calls,
            self.mock_message_response,
        ]
        self.tool_manager.handle_tool_calls.return_value = [tool_response_mock]

        sut = JAImsAgent(
            llm_interface=self.llm_interface,
            tool_manager=self.tool_manager,
            tools=self.mock_tools,
        )

        response = sut.message([first_message_mock])
        self.assertEqual(response, "World")

        self.tool_manager.handle_tool_calls.assert_called_once_with(
            sut, self.mock_message_response_tool_calls.tool_calls, self.mock_tools
        )

    def test_tools_are_overridden_when_passed_trough_run_method(self):
        self.llm_interface.call.return_value = self.mock_message_response_tool_calls
        self.tool_manager.handle_tool_calls.return_value = []

        sut = JAImsAgent(
            llm_interface=self.llm_interface,
            tool_manager=self.tool_manager,
            tools=self.mock_tools,
        )

        override_tools = MagicMock()

        response = sut.message([self.mock_message], [override_tools])

        self.llm_interface.call.assert_called_once_with(
            [self.mock_message], [override_tools], tool_constraints=None
        )

        self.tool_manager.handle_tool_calls.assert_called_once_with(
            sut, self.mock_message_response_tool_calls.tool_calls, [override_tools]
        )


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional

from jaims.entities import JAImsMessage
from jaims.interfaces import JAImsHistoryManager, JAImsHistoryOptimizer


class JAImsLastNHistoryOptimizer(JAImsHistoryOptimizer):
    """
    A history optimizer that keeps only the last N messages in the history.
    """

    def __init__(self, last_n: int):
        self.__last_n = last_n

    def optimize_history(self, messages: List[JAImsMessage]) -> List[JAImsMessage]:
        """
        Optimizes the history by keeping only the last N messages.

        Args:
            messages (List[JAImsMessage]): The list of messages to optimize.

        Returns:
            List[JAImsMessage]: The optimized list of messages.
        """
        return messages[-self.__last_n :] if self.__last_n > 0 else []


class JAImsDefaultHistoryManager(JAImsHistoryManager):
    """
    A class that handles and manages the history of JAIms messages when used as a chat-agent.

    Use the attribtutes leading_prompts and trailing_prompts to add prompts that will be added to the beginning and end of the history respectively.

    The method get_messages retrieves the messages in the history that will be sent to the chat agent, optionally optimized by the history optimizer.

    Attributes:
        history (Optional[List[JAImsMessage]]): The list of JAIms messages representing the history.
        leading_prompts (Optional[List[JAImsMessage]]): The list of JAIms messages representing the leading prompts.
        trailing_prompts (Optional[List[JAImsMessage]]): The list of JAIms messages representing the trailing prompts.
        history_optimizer (Optional[JAImsHistoryOptimizer]): The history optimizer used to optimize the history.

    Methods:
        add_messages(messages: List[JAImsMessage]): Adds a list of JAIms messages to the history.
        get_messages(): Retrieves the messages in the history, optionally optimized by the history optimizer.
        clear_history(): Clears the history.
        get_history(): Returns the entire history.
    """

    def __init__(
        self,
        history: Optional[List[JAImsMessage]] = None,
        leading_prompts: Optional[List[JAImsMessage]] = None,
        trailing_prompts: Optional[List[JAImsMessage]] = None,
        history_optimizer: Optional[JAImsHistoryOptimizer] = None,
    ):
        self.__history = history or []
        self.__leading_prompts = leading_prompts or []
        self.__trailing_prompts = trailing_prompts or []
        self.__history_optimizer = history_optimizer

    def add_messages(self, messages: List[JAImsMessage]):
        """
        Adds a list of JAIms messages to the history.

        Args:
            messages (List[JAImsMessage]): The list of JAIms messages to be added to the history.
        """
        self.__history.extend(messages)

    def get_messages(self):
        """
        Retrieves the messages in the history, along with the leading and trailing prompts, optionally optimized by the history optimizer.

        Returns:
            List[JAImsMessage]: The list of JAIms messages in the history.
        """
        if self.__history_optimizer:
            return (
                self.__leading_prompts
                + self.__history_optimizer.optimize_history(self.__history)
                + self.__trailing_prompts
            )

        return self.__leading_prompts + self.__history + self.__trailing_prompts

    def clear_history(self):
        """
        Clears the history. Does not clear the leading and trailing prompts.
        """
        self.__history = []

    def get_history(self):
        """
        Returns the entire message history.

        Returns:
            List[JAImsMessage]: The list of JAIms messages representing the entire history.
        """
        return self.__history



from enum import Enum
import random
from typing import Any, Callable
from ...entities import JAImsOptions, JAImsMaxRetriesExceeded
import time
import logging


class ErrorHandlingMethod(Enum):
    FAIL = "fail"
    RETRY = "retry"
    EXPONENTIAL_BACKOFF = "exponential_backoff"


def call_with_exponential_backoff(
    operation: Callable[[], Any],
    error_handler: Callable[[Exception], ErrorHandlingMethod],
    options: JAImsOptions,
) -> Any:
    retries = 0
    sleep_time = options.retry_delay
    backoff_time = options.exponential_delay
    logger = logging.getLogger(__name__)
    last_exception = None

    while retries <= options.max_retries:
        try:
            return operation()
        except Exception as error:
            last_exception = error
            logger.error(f"Client Error:\n{error}\n")
            error_handling_method = error_handler(error)

            if error_handling_method == ErrorHandlingMethod.FAIL:
                raise error

            if error_handling_method == ErrorHandlingMethod.RETRY:
                sleep_time = options.retry_delay

            elif error_handling_method == ErrorHandlingMethod.EXPONENTIAL_BACKOFF:
                logger.warning(f"Performing exponential backoff")
                jitter = 1 + options.jitter * random.random()
                backoff_time = backoff_time * options.exponential_base * jitter

                if (
                    options.exponential_cap is not None
                    and backoff_time > options.exponential_cap
                ):
                    backoff_time = options.exponential_cap * jitter

                sleep_time = backoff_time

            retries += 1
            logger.warning(
                f"Attempting {retries} of {options.max_retries}. in {sleep_time} seconds"
            )
            time.sleep(sleep_time)

    logger.error(f"Max retries exceeded! Operation failed {options.max_retries} times.")
    raise JAImsMaxRetriesExceeded(options.max_retries, last_exception)



from .adapter import (
    JAImsOpenaiKWArgs,
    JAImsTokenHistoryOptimizer,
    OpenAITransactionStorageInterface,
    JAImsOpenaiAdapter,
)

from .factory import create_jaims_openai

__all__ = [
    "JAImsOpenaiKWArgs",
    "create_jaims_openai",
    "JAImsTokenHistoryOptimizer",
    "OpenAITransactionStorageInterface",
    "JAImsOpenaiAdapter",
]



from setuptools import setup, find_packages

# Read the main requirements
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

# Read OpenAI specific requirements
with open("requirements-openai.txt") as f:
    requirements_openai = f.read().splitlines()

# Read Google Cloud AI specific requirements
with open("requirements-googleai.txt") as f:
    requirements_google_ai = f.read().splitlines()

setup(
    name="jaims-py",
    version="2.0.0-beta.9",
    packages=find_packages(),
    description="A Python package for creating LLM powered, agentic, platform agnostic software.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Marco Musella",
    url="https://github.com/dev-mush/jaims-py",
    license="MIT",
    keywords="An extensible library to create LLM Agents and LLM based applications.",
    install_requires=requirements,
    extras_require={
        "openai": requirements_openai,
        "google": requirements_google_ai,
    },
)



from typing import List, Optional
from google.generativeai.types import generation_types

from ...agent import JAImsAgent
from ...entities import JAImsFunctionTool
from ...interfaces import JAImsHistoryManager, JAImsToolManager
from .adapter import JAImsGoogleGenerativeAIAdapter


def create_jaims_gemini(
    model: str = "gemini-1.5-pro",
    api_key: Optional[str] = None,
    generation_config: Optional[generation_types.GenerationConfigType] = None,
    history_manager: Optional[JAImsHistoryManager] = None,
    tool_manager: Optional[JAImsToolManager] = None,
    tools: Optional[List[JAImsFunctionTool]] = None,
) -> JAImsAgent:

    adapter = JAImsGoogleGenerativeAIAdapter(
        api_key=api_key,
        model=model,
        generation_config=generation_config,
    )

    agent = JAImsAgent(
        llm_interface=adapter,
        history_manager=history_manager,
        tool_manager=tool_manager,
        tools=tools,
    )

    return agent



from typing import Dict, List, Optional, Union
from ...agent import JAImsAgent
from ...entities import JAImsFunctionTool, JAImsOptions, JAImsLLMConfig
from ...interfaces import JAImsHistoryManager, JAImsToolManager
from .adapter import (
    JAImsOpenaiAdapter,
    JAImsOpenaiKWArgs,
    OpenAITransactionStorageInterface,
)


def create_jaims_openai(
    api_key: Optional[str] = None,
    options: Optional[JAImsOptions] = None,
    kwargs: Optional[Union[JAImsOpenaiKWArgs, Dict]] = None,
    transaction_storage: Optional[OpenAITransactionStorageInterface] = None,
    history_manager: Optional[JAImsHistoryManager] = None,
    tool_manager: Optional[JAImsToolManager] = None,
    tools: Optional[List[JAImsFunctionTool]] = None,
) -> JAImsAgent:

    adapter = JAImsOpenaiAdapter(
        api_key=api_key,
        options=options,
        kwargs=kwargs,
        transaction_storage=transaction_storage,
    )

    agent = JAImsAgent(
        llm_interface=adapter,
        history_manager=history_manager,
        tool_manager=tool_manager,
        tools=tools,
    )

    return agent



from __future__ import annotations
from abc import ABC, abstractmethod
import json
from math import ceil
from typing import Generator, Union
import openai
from openai import OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import ChoiceDelta, ChoiceDeltaToolCall
from openai import Stream
import tiktoken
from typing import List, Optional, Dict
from PIL import Image

from ...interfaces import (
    JAImsLLMInterface,
    JAImsHistoryOptimizer,
)
from ...entities import (
    JAImsImageContent,
    JAImsContentType,
    JAImsMessage,
    JAImsStreamingMessage,
    JAImsToolCall,
    JAImsFunctionTool,
    JAImsMessageRole,
    JAImsOptions,
)
from ..shared.image_utilities import image_to_b64
from ..shared.exponential_backoff_operation import (
    call_with_exponential_backoff,
    ErrorHandlingMethod,
)

import os
from copy import deepcopy

# ---------------------
# openai / LLM modeling
# ---------------------


class JAImsOpenaiKWArgs:
    """
    Represents the keyword arguments for the JAIms OpenAI wrapper.
    This class entirely mirrors the openai API parameters, so refer to it for documentation.
    (https://platform.openai.com/docs/api-reference/chat/create).

    Args:
        model (str, optional): The OpenAI model to use. Defaults to gpt-3.5-turbo.
        messages (List[dict], optional): The list of messages for the chat completion. Defaults to an empty list, it is automatically populated by the run method so it is not necessary to pass them. If passed, they will always be appended to the messages passed in the run method.
        max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to 500.
        stream (bool, optional): Whether to use streaming for the API call. Defaults to False.
        temperature (float, optional): The temperature for generating creative text. Defaults to 0.0.
        top_p (Optional[int], optional): The top-p value for nucleus sampling. Defaults to None.
        n (int, optional): The number of responses to generate. Defaults to 1.
        seed (Optional[int], optional): The seed to be passed to openai to have more consistent outputs. Defaults to None.
        frequency_penalty (float, optional): The frequency penalty for avoiding repetitive responses. Defaults to 0.0.
        presence_penalty (float, optional): The presence penalty for encouraging diverse responses. Defaults to 0.0.
        logit_bias (Optional[Dict[str, float]], optional): The logit bias for influencing the model's output. Defaults to None.
        response_format (Optional[Dict], optional): The format for the generated response. Defaults to None.
        stop (Union[Optional[str], Optional[List[str]]], optional): The stop condition for the generated response. Defaults to None.
        tool_choice (Union[str, Dict], optional): The choice of tool to use. Defaults to "auto".
        tools (Optional[List[JAImsFunctionToolWrapper]], optional): The list of function tool wrappers to use. Defaults to None.
    """

    def __init__(
        self,
        model: str = "gpt-3.5-turbo",
        messages: List[dict] = [],
        max_tokens: int = 1024,
        stream: bool = False,
        temperature: float = 0.0,
        top_p: Optional[int] = None,
        n: int = 1,
        seed: Optional[int] = None,
        frequency_penalty: float = 0.0,
        presence_penalty: float = 0.0,
        logit_bias: Optional[Dict[str, float]] = None,
        response_format: Optional[Dict] = None,
        stop: Union[Optional[str], Optional[List[str]]] = None,
        tool_choice: Union[str, Dict] = "auto",
        tools: Optional[List[Dict]] = None,
    ):
        self.model = model
        self.messages = messages
        self.max_tokens = max_tokens
        self.stream = stream
        self.temperature = temperature
        self.top_p = top_p
        self.n = n
        self.seed = seed
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.logit_bias = logit_bias
        self.response_format = response_format
        self.stop = stop
        self.tool_choice = tool_choice
        self.tools = tools

    def to_dict(self):
        kwargs = {
            "model": self.model,
            "temperature": self.temperature,
            "n": self.n,
            "stream": self.stream,
            "messages": self.messages,
            "top_p": self.top_p,
            "max_tokens": self.max_tokens,
            "seed": self.seed,
            "tools": self.tools,
            "tool_choice": self.tool_choice,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "response_format": self.response_format,
            "stop": self.stop,
        }

        # Remove None values
        kwargs = {key: value for key, value in kwargs.items() if value is not None}

        return kwargs

    @staticmethod
    def from_dict(kwargs: dict) -> JAImsOpenaiKWArgs:
        return JAImsOpenaiKWArgs(
            model=kwargs.get("model", "gpt-3.5-turbo"),
            messages=kwargs.get("messages", []),
            max_tokens=kwargs.get("max_tokens", 1024),
            stream=kwargs.get("stream", False),
            temperature=kwargs.get("temperature", 0.0),
            top_p=kwargs.get("top_p", None),
            n=kwargs.get("n", 1),
            seed=kwargs.get("seed", None),
            frequency_penalty=kwargs.get("frequency_penalty", 0.0),
            presence_penalty=kwargs.get("presence_penalty", 0.0),
            logit_bias=kwargs.get("logit_bias", None),
            response_format=kwargs.get("response_format", None),
            stop=kwargs.get("stop", None),
            tool_choice=kwargs.get("tool_choice", "auto"),
            tools=kwargs.get("tools", None),
        )

    def copy_with_overrides(
        self,
        model: Optional[str] = None,
        messages: Optional[List[dict]] = None,
        max_tokens: Optional[int] = None,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[int] = None,
        n: Optional[int] = None,
        seed: Optional[int] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, float]] = None,
        response_format: Optional[Dict] = None,
        stop: Optional[Union[str, List[str]]] = None,
        tool_choice: Optional[Union[str, Dict]] = None,
        tools: Optional[List[Dict]] = None,
    ) -> JAImsOpenaiKWArgs:
        """
        Returns a new JAImsOpenaiKWArgs instance with the passed kwargs overridden.
        """
        return JAImsOpenaiKWArgs(
            model=model if model else self.model,
            messages=messages if messages else self.messages,
            max_tokens=max_tokens if max_tokens else self.max_tokens,
            stream=stream if stream else self.stream,
            temperature=temperature if temperature else self.temperature,
            top_p=top_p if top_p else self.top_p,
            n=n if n else self.n,
            seed=seed if seed else self.seed,
            frequency_penalty=(
                frequency_penalty if frequency_penalty else self.frequency_penalty
            ),
            presence_penalty=(
                presence_penalty if presence_penalty else self.presence_penalty
            ),
            logit_bias=logit_bias if logit_bias else self.logit_bias,
            response_format=(
                response_format if response_format else self.response_format
            ),
            stop=stop if stop else self.stop,
            tool_choice=tool_choice if tool_choice else self.tool_choice,
            tools=tools if tools else self.tools,
        )


class JAImsTokenHistoryOptimizer(JAImsHistoryOptimizer):
    def __init__(
        self,
        options: JAImsOptions,
        history_max_tokens: int,
        model: str,
    ):
        self.options = options
        self.history_max_tokens = history_max_tokens
        self.model = model

    def optimize_history(self, messages: List[JAImsMessage]) -> List:

        # Copying the whole history to avoid altering the original one
        buffer = messages.copy()

        # calculate the tokens for the compound history
        messages_tokens = self.__tokens_from_messages(buffer, self.model)

        while messages_tokens > self.history_max_tokens:
            if not buffer:
                raise Exception(
                    f"Unable to fit messages with current max tokens {self.history_max_tokens}."
                )
            # Popping the first (oldest) message from the chat history between the user and agent
            buffer.pop(0)
            # Recalculating the tokens for the compound history
            messages_tokens = self.__tokens_from_messages(buffer, self.model)

        return buffer

    def __estimate_token_count(self, string: str, model: str) -> int:
        """Returns the number of tokens in a text string."""

        encoding = tiktoken.encoding_for_model(model)
        num_tokens = len(encoding.encode(string))
        return num_tokens

    def __estimate_image_tokens_count(self, width: int, height: int):
        h = ceil(height / 512)
        w = ceil(width / 512)
        n = w * h
        total = 85 + 170 * n
        return total

    def __tokens_from_messages(self, messages: List[JAImsMessage], model):
        """Returns the number of tokens used by a list of messages."""

        images = []
        parsed = []
        for message in messages:
            if message.contents:
                for item in message.contents:
                    if isinstance(item, str):
                        parsed.append(item)
                    elif isinstance(item, Image.Image):
                        images.append(item)
                    else:
                        raise Exception(f"Unsupported content type: {type(item)}")

            if message.tool_calls:
                for tool_call in message.tool_calls:
                    parsed.append(tool_call.tool_name + json.dumps(tool_call.tool_args))

            if message.tool_response:
                parsed.append(message.tool_response.response)

        image_tokens = 0
        for image in images:
            width, height = image.size
            image_tokens += self.__estimate_image_tokens_count(width, height)

        return self.__estimate_token_count(json.dumps(parsed), model) + image_tokens


class OpenAITransactionStorageInterface(ABC):
    """
    Interface for storing LLM transactions.
    Override this class to implement your own storage, to store a pair of LLM request and response payloads.
    """

    @abstractmethod
    def store_transaction(self, request: dict, response: dict):
        pass


class JAImsOpenaiAdapter(JAImsLLMInterface):
    """
    The JAIms OpenAI adapter.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        options: Optional[JAImsOptions] = None,
        kwargs: Optional[Union[JAImsOpenaiKWArgs, Dict]] = None,
        transaction_storage: Optional[OpenAITransactionStorageInterface] = None,
    ):
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        if not self.api_key:
            raise Exception("OpenAI API key not provided.")

        self.options = options or JAImsOptions()
        self.kwargs = kwargs or JAImsOpenaiKWArgs()
        self.transaction_storage = transaction_storage

    def __get_args(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
        stream: bool = False,
    ):

        if isinstance(self.kwargs, JAImsOpenaiKWArgs):
            args = self.kwargs.to_dict()
        else:
            args = deepcopy(self.kwargs)

        openai_messages = self.__jaims_messages_to_openai(messages)
        args["messages"] = openai_messages
        args["stream"] = stream

        if tools:
            openai_tools = self.__jaims_tools_to_openai(tools)

            tool_choice = "auto"
            if tool_constraints:
                if len(tool_constraints) > 1:
                    raise ValueError(
                        "Only one tool choice is allowed when using the OpenAI API."
                    )

                tool_choice = {
                    "type": "function",
                    "function": {
                        "name": tool_constraints[0],
                    },
                }

            args["tools"] = openai_tools
            args["tool_choice"] = tool_choice

        return args

    def call(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
    ) -> JAImsMessage:
        args = self.__get_args(messages, tools, tool_constraints)
        response = self.___get_openai_response(args, self.options)
        assert isinstance(response, ChatCompletion)
        if self.transaction_storage:
            self.transaction_storage.store_transaction(
                request=args,
                response=response.model_dump(exclude_none=True),
            )

        return self.__openai_chat_completion_to_jaims_message(response)

    def call_streaming(
        self,
        messages: List[JAImsMessage],
        tools: List[JAImsFunctionTool],
        tool_constraints: Optional[List[str]] = None,
    ) -> Generator[JAImsStreamingMessage, None, None]:
        args = self.__get_args(
            messages, tools, stream=True, tool_constraints=tool_constraints
        )
        response = self.___get_openai_response(args, self.options)
        assert isinstance(response, Stream)

        accumulated_delta = None
        for completion_chunk in response:
            accumulated_delta = self.__accumulate_choice_delta(
                accumulated_delta, completion_chunk.choices[0].delta
            )
            yield self.__openai_chat_completion_choice_delta_to_jaims_message(
                accumulated_delta, completion_chunk
            )

        if self.transaction_storage and accumulated_delta:
            self.transaction_storage.store_transaction(
                request=args,
                response=accumulated_delta.model_dump(exclude_none=True),
            )

    def __jaims_messages_to_openai(self, messages: List[JAImsMessage]) -> List[dict]:

        def format_contents(contents: List[JAImsContentType]):
            if len(contents) == 1 and isinstance(contents[0], str):
                return contents[0]
            else:
                raw_contents = []
                for c in contents:
                    if isinstance(c, str):
                        raw_contents.append({"type": "text", "text": c})
                    elif isinstance(c, JAImsImageContent):
                        url = c.image
                        if isinstance(c.image, Image.Image):
                            mime, b64 = image_to_b64(c.image)
                            url = f"data:{mime};base64,{b64}"

                        raw_contents.append(
                            {
                                "type": "image_url",
                                "image_url": {"url": url},
                            }
                        )

                    else:
                        raise Exception(f"Unsupported content type: {type(c)}")

                return raw_contents

        def format_tool_calls(tool_calls: List[JAImsToolCall]):
            raw_tool_calls = []
            for tc in tool_calls:
                raw_tool_call = {
                    "type": "function",
                    "id": tc.id,
                    "function": {
                        "name": tc.tool_name,
                        "arguments": json.dumps(tc.tool_args),
                    },
                }
                raw_tool_calls.append(raw_tool_call)
            return raw_tool_calls

        raw_messages = []
        for m in messages:
            if m.tool_response:
                raw_messages.append(
                    {
                        "role": "tool",
                        "name": m.tool_response.tool_name,
                        "tool_call_id": m.tool_response.tool_call_id,
                        "content": json.dumps(m.tool_response.response),
                    }
                )
                continue

            if m.tool_calls:
                raw_messages.append(
                    {
                        "role": "assistant",
                        "tool_calls": format_tool_calls(m.tool_calls),
                    }
                )
                continue

            raw_message = {}
            raw_message["role"] = m.role.value

            if m.contents:
                raw_message["content"] = format_contents(m.contents)

            raw_messages.append(raw_message)

        return raw_messages

    def __jaims_tools_to_openai(self, tools: List[JAImsFunctionTool]) -> List[dict]:
        raw_tools = []
        for t in tools:
            tool_raw_dict = {
                "type": "function",
                "function": {
                    "name": t.descriptor.name,
                    "description": t.descriptor.description,
                    "parameters": t.descriptor.json_schema(),
                },
            }
            raw_tools.append(tool_raw_dict)

        return raw_tools

    def __openai_chat_completion_to_jaims_message(
        self, completion: ChatCompletion
    ) -> JAImsMessage:
        if len(completion.choices) == 0:
            raise Exception("OpenAI returned an empty response.")

        message = completion.choices[0].message
        role = JAImsMessageRole(message.role)
        tool_calls = None
        if message.tool_calls:
            tool_calls = [
                JAImsToolCall(
                    id=tc.id,
                    tool_name=tc.function.name,
                    tool_args=json.loads(tc.function.arguments),
                )
                for tc in message.tool_calls
            ]

        return JAImsMessage(
            role=role,
            contents=[message.content] if message.content else None,
            tool_calls=tool_calls,
            raw=message,
        )

    def __openai_chat_completion_choice_delta_to_jaims_message(
        self, accumulated_choice_delta: ChoiceDelta, current_chunk: ChatCompletionChunk
    ) -> JAImsStreamingMessage:

        current_choice = current_chunk.choices[0]
        role = (
            JAImsMessageRole(current_choice.delta.role)
            if current_choice.delta.role
            else accumulated_choice_delta.role
        )
        textDelta = current_choice.delta.content
        contents: Optional[List[JAImsContentType]] = None
        function_tool_calls = None

        role = JAImsMessageRole(accumulated_choice_delta.role)
        if accumulated_choice_delta.content:
            contents = [
                accumulated_choice_delta.content,
            ]

        if current_choice.finish_reason and accumulated_choice_delta.tool_calls:
            function_tool_calls = []
            for tc in accumulated_choice_delta.tool_calls:
                if tc.function:
                    function_tool_calls.append(
                        JAImsToolCall(
                            id=tc.id or "",
                            tool_name=tc.function.name or "",
                            tool_args=(
                                json.loads(tc.function.arguments)
                                if tc.function.arguments
                                else None
                            ),
                        )
                    )

        return JAImsStreamingMessage(
            message=JAImsMessage(
                role=role,
                contents=contents,
                tool_calls=function_tool_calls,
                raw=accumulated_choice_delta,
            ),
            textDelta=textDelta,
        )

    def ___get_openai_response(
        self,
        openai_kw_args: dict,
        call_options: JAImsOptions,
    ) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:

        def handle_openai_error(error) -> ErrorHandlingMethod:
            # errors are handled according to the guidelines here: https://platform.openai.com/docs/guides/error-codes/api-errors (dated 03/10/2023)
            # this map indexes all the error that require a retry or an exponential backoff, every other error is a fail
            error_handling_map = {
                openai.RateLimitError: ErrorHandlingMethod.EXPONENTIAL_BACKOFF,
                openai.InternalServerError: ErrorHandlingMethod.RETRY,
                openai.APITimeoutError: ErrorHandlingMethod.RETRY,
            }

            for error_type, error_handling_method in error_handling_map.items():
                if isinstance(error, error_type):
                    return error_handling_method

            return ErrorHandlingMethod.FAIL

        def openai_api_call():
            client = OpenAI(api_key=self.api_key)
            kwargs = openai_kw_args
            response = client.chat.completions.create(
                **kwargs,
            )

            return response

        return call_with_exponential_backoff(
            openai_api_call,
            handle_openai_error,
            call_options,
        )

    def __merge_tool_calls(
        self,
        existing_tool_calls: Optional[List[ChoiceDeltaToolCall]],
        new_tool_calls_delta: List[ChoiceDeltaToolCall],
    ):
        if not existing_tool_calls:
            return new_tool_calls_delta

        new_tool_calls = existing_tool_calls[:]
        for new_call_delta in new_tool_calls_delta:
            # check the tall call is already being streamed
            existing_call = next(
                (item for item in new_tool_calls if item.index == new_call_delta.index),
                None,
            )
            # new tool call, add it to the list
            if not existing_call:
                new_tool_calls.append(new_call_delta)

            # existing tool call, update it
            else:
                # update tool type
                if (
                    existing_call.type != new_call_delta.type
                    and new_call_delta.type is not None
                ):
                    existing_call.type = new_call_delta.type

                # update tool id
                if (
                    existing_call.id != new_call_delta.id
                    and new_call_delta.id is not None
                ):
                    existing_call.id = new_call_delta.id

                # update function
                if new_call_delta.function:
                    if existing_call.function is None:
                        existing_call.function = new_call_delta.function
                    else:
                        # update function name
                        if (
                            existing_call.function.name != new_call_delta.function.name
                            and new_call_delta.function.name is not None
                        ):
                            existing_call.function.name = new_call_delta.function.name

                        # update function args
                        existing_call.function.arguments = (
                            existing_call.function.arguments or ""
                        ) + (new_call_delta.function.arguments or "")

        return new_tool_calls

    def __accumulate_choice_delta(
        self, accumulator: Optional[ChoiceDelta], new_delta: ChoiceDelta
    ) -> ChoiceDelta:
        if accumulator is None:
            return new_delta

        if new_delta.content:
            accumulator.content = (accumulator.content or "") + new_delta.content
        if new_delta.role:
            accumulator.role = new_delta.role
        if new_delta.tool_calls:
            accumulator.tool_calls = self.__merge_tool_calls(
                accumulator.tool_calls, new_delta.tool_calls
            )

        return accumulator



from __future__ import annotations
from typing import Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    from .agent import JAImsAgent

from jaims.entities import JAImsOptions, JAImsLLMConfig
from jaims.interfaces import JAImsHistoryManager, JAImsToolManager, JAImsFunctionTool


def openai_factory(
    model: str,
    api_key: Optional[str] = None,
    options: Optional[JAImsOptions] = None,
    config: Optional[JAImsLLMConfig] = None,
    history_manager: Optional[JAImsHistoryManager] = None,
    tool_manager: Optional[JAImsToolManager] = None,
    tools: Optional[List[JAImsFunctionTool]] = None,
) -> JAImsAgent:
    """
    Factory function to create an instance of JAImsAgent using OpenAI as the underlying model.

    Args:
        model (str): The name or identifier of the OpenAI model to use.
        api_key (Optional[str]): The API key for accessing the OpenAI service. Defaults to None.
        options (Optional[JAImsOptions]): Additional options for configuring the JAImsAgent. Defaults to None.
        config (Optional[JAImsLLMConfig]): Configuration options specific to the OpenAI language model. Defaults to None.
        history_manager (Optional[JAImsHistoryManager]): The history manager to use for managing conversation history. Defaults to None.
        tool_manager (Optional[JAImsToolManager]): The tool manager to use for managing JAImsFunctionTools. Defaults to None.
        tools (Optional[List[JAImsFunctionTool]]): The list of JAImsFunctionTools to use. Defaults to None.

    Returns:
        JAImsAgent: An instance of JAImsAgent configured with the OpenAI model.

    """
    from .adapters.openai_adapter import create_jaims_openai
    from .adapters.openai_adapter import JAImsOpenaiKWArgs

    config = config or JAImsLLMConfig()
    options = options or JAImsOptions()
    tool_choice = None

    kwargs = JAImsOpenaiKWArgs().copy_with_overrides(
        model=model,
        temperature=config.temperature,
        max_tokens=config.max_tokens,
        response_format=config.response_format,
        tool_choice=tool_choice,
    )

    return create_jaims_openai(
        api_key=api_key,
        options=options,
        kwargs=kwargs,
        history_manager=history_manager,
        tool_manager=tool_manager,
        tools=tools,
    )


def google_factory(
    model: str,
    api_key: Optional[str] = None,
    options: Optional[JAImsOptions] = None,
    config: Optional[JAImsLLMConfig] = None,
    history_manager: Optional[JAImsHistoryManager] = None,
    tool_manager: Optional[JAImsToolManager] = None,
    tools: Optional[List[JAImsFunctionTool]] = None,
) -> JAImsAgent:
    """
    Factory function to create an instance of JAImsAgent using Google Gemini as the underlying model.

    Args:
        model (str): The name or ID of the Google model to use.
        api_key (Optional[str]): The API key for accessing the Google model (default: None).
        options (Optional[JAImsOptions]): Additional options for the JAImsAgent (default: None).
        config (Optional[JAImsLLMConfig]): Configuration for the JAImsLLM model (default: None).
        history_manager (Optional[JAImsHistoryManager]): History manager for the JAImsAgent (default: None).
        tool_manager (Optional[JAImsToolManager]): Tool manager for the JAImsAgent (default: None).
        tools (Optional[List[JAImsFunctionTool]]): List of function tools for the JAImsAgent (default: None).

    Returns:
        JAImsAgent: An instance of JAImsAgent configured for Google models.
    """

    from .adapters.google_generative_ai_adapter.factory import (
        create_jaims_gemini,
        generation_types,
    )

    config = config or JAImsLLMConfig()
    options = options or JAImsOptions()

    generation_config = generation_types.GenerationConfig(
        temperature=config.temperature,
        max_output_tokens=config.max_tokens,
        response_schema=config.response_format,
    )

    return create_jaims_gemini(
        api_key=api_key,
        model=model,
        generation_config=generation_config,
        history_manager=history_manager,
        tool_manager=tool_manager,
        tools=tools,
    )
